{{
'-------------------------------------------------------------------------------
'
' LMM_Debug - Implements a debug kernel for the Large Memory Model Kernel 
'             for use by the Catalina Code Generator backend for LCC
'
' This kernel has support for the POD debugger. To make space for POD, the
' floating point routines are sent off to the Float32A plugin instead of
' being processed internally by the Kernel. This makes this LMM slower, but
' gives more space.
'
' The following symbols affect the plugins loaded - to
' define them, use the -D command line option (e.g. -DVGA)
' except for the library symbols, which are defined when the
' appropriate library is included (e.g. -lma, -lmb)
'
'    NO_FP          - do not load any FP plugins
'    NO_FLOAT       - do not load any FP Plugins (new in 2.8)
'    NO_HMI         - do not load any HMI plugin
'    NO_MOUSE       - do not load/start mouse driver
'    NO_KEYBOARD    - do not load/start keyboard driver
'    NO_SCREEN      - do not load/start screen driver
'    HIRES_VGA      - load a HiRes VGA driver
'    LORES_VGA      - load a LoRes VGA driver
'    VGA            -   "  "   "    "    "
'    HIRES_TV       - load a HiRes TV driver 
'    LORES_TV       - load a LoRes TV driver 
'    TV             -   "  "   "    "    "    
'    NO_INTERLACE   - use non-interlace mode (TV drivers only)
'    NTSC           - use NTSC mode (TV drivers only)    
'    PC             - use a PC HMI plugin (screen/kbd only)
'    PROPTERMINAL   - use a PropTerminal HMI plugin
'    CLOCK          - load a Real-Time Clock plugin
'    libma          - generated automatically by -lma option
'    libmb          - generated automatically by -lmb option
'    libcx          - generated automatically by -lcx option
'    PROXY_KEYBOARD - use a proxy for the keyboard driver
'    PROXY_MOUSE    - use a proxy for the mouse driver
'    PROXY_SCREEN   - use a proxy for the screen driver
'    PROXY_SD       - use a proxy for the SD plugin
'
'    PROXY_HMI      - (internal) load the proxy HMI plugin
'    PROXY_SIO      - (internal) load the SIO plugin for use by the proxy plugin
'
'    RESERVE_COG    - Load the reserve cog
'                 
' Version 1.0 - initial version - by Ross Higson
' Version 2.0 - use #defines to determine plugins to load 
' Version 2.2 - add proxy plugin support
' Version 2.5 - Add Pullmoll's new f_d32s routine, and move debug vectors 
' Version 2.8 - Added support for reserve cog.
'               Simplified object selection logic.
'               Added NO_FLOAT (equivalent to NO_FP)
'
' This program incorporates software derived from
'    POD (Propeller On-Chip Debugger), (C) 2007 Thomas Kliche
'
'------------------------------------------------------------------------------
'
'    Copyright 2009 Ross Higson
'
'    The portion of this file identified as the LMM Kernel is part of the 
'    Catalina Target Package.
'
'    The Catalina Target Package is free software: you can redistribute 
'    it and/or modify it under the terms of the GNU Lesser General Public 
'    License as published by the Free Software Foundation, either version 
'    3 of the License, or (at your option) any later version.
'
'    The Catalina Target Package is distributed in the hope that it will
'    be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
'    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
'    See the GNU Lesser General Public License for more details.
'
'    You should have received a copy of the GNU Lesser General Public 
'    License along with the Catalina Target Package.  If not, see 
'    <http://www.gnu.org/licenses/>.
'
'------------------------------------------------------------------------------
}}
{{
*****************************
* Demonstration code using  *
* debug kernel for assembly *
* code on-chip debugging    *
*  (C) 2007 Thomas Kliche   *
*****************************

Revision History
----------------
 1.1 on 29 March 2007
     - all fixed made in AsmDebug
     - added demo code containing jmpret instruction

 1.0 on 18 March 2007
     - first public version

...................................................................................................

This is a debug kernel which provides some generic functionality that you can use
to debug your assembly code running in a Cog.

---------------------------------------------------------------------------------------------------
}}
CON

  FLAG_C          = %01                                 'bit position for C flag
  FLAG_NZ         = %10                                 'bit position for inverted Z flag

  'commands for communication with debug kernel
  CMD_RDMAINDATA  = 1                                   'read data from main memory
  CMD_WRMAINDATA  = 2                                   'write data to main memory
  CMD_RDCOGDATA   = 3                                   'read data from Cog memory
  CMD_WRCOGDATA   = 4                                   'write data to Cog memory
  CMD_RDFLAGS     = 5                                   'read flags

CON
'
LMM_FLA      = Common#LMM_FLA
'
FLOAT_XFER   = Common#FLOAT_XFER
'
' Float32 constants.
'
SignFlag      = $1
ZeroFlag      = $2
NaNFlag       = $8
'
FAddCmd       = 1 << 24       ' Send this command to Float32_A
FSubCmd       = 2 << 24       ' Send this command to Float32_A
FMulCmd       = 3 << 24       ' Send this command to Float32_A
FDivCmd       = 4 << 24       ' Send this command to Float32_A
FFloatCmd     = 5 << 24       ' Implemented in Float32_A and Float32_B
FTruncCmd     = 6 << 24       ' Implemented in Float32_A and Float32_B
'
' Set up Proxy and HMI symbols and constants:
'
#include "Constants.inc"
'
VAR

  long  DebugCog

OBJ
'
' The following objects MUST be loaded ...
'
  Common   : "Catalina_Common"                          ' Common Definitions
'
' Include the HMI selection logic and load the appropriate objects:
'
#include "HMI.inc"
'
' Include the Floating Point selection logic and load the appropriate objects:
'
#include "FP.inc"
'
' Include selection logic for other plugins and load the appropriate objects:
'
#include "Other.inc"
'
PRI PreStart

  Common.InitializeRegistry

PRI StartCog : cog

  ' start the LMM (it won't run till we give it the base address)

  cog := cognew(@debugEntry, Common#REGISTRY) 
  
PRI PostStart (cog, base) | ok, ALLOC, HMI_DATA, CGI_DATA, RSV_DATA, IO_BLOCK, Proxy_Lock, Server_CPU, s

#ifdef RESERVE_COG
  ' Point to start of File buffers (we must allocate data blocks down from here)
  ALLOC := Common#FLIST_BUFF
#else  
  ' Point to start of Registry (we must allocate data blocks down from here)
  ALLOC := Common#REQUESTS
#endif

  ok         := TRUE
  Proxy_Lock := -1

#ifdef RESERVE_COG
  ' allocate Hub RAM for reserved cog
  ' (the reserve cog is always allocated space and started first)
  RSV_DATA := ALLOC - RSV#BYTE_SIZE
  ALLOC    := RSV_DATA
  if ok
    ' start the reserved cog
    ok := RSV.Start(RSV_DATA)
#endif

#ifdef NO_ARGS
  CogStore.Stop
#else  
  s := CogStore.Size
  if (s => 0)
    ALLOC := ALLOC - s*4
  CogStore.Setup(ALLOC)
  CogStore.Stop
#endif  

#ifdef PROXY_SIO
  ' Figure out the server CPU we should use as proxy
  ' (works on Morpheus & TriBladeProp, may not work on others):
#ifdef CPU_1
  Server_CPU  := 2
#elseifdef CPU_2
  Server_CPU  := 1
#elseifdef CPU_3
  Server_CPU  := 2
#endif
  ' allocate a lock to serialize proxy access
  Proxy_Lock := locknew
  ' allocate a data block for SIO
  IO_BLOCK := ALLOC - SIO#BLOCK_SIZE
  ALLOC    := IO_BLOCK
  ' start the SIO plugin
  if ok
    ok := SIO.Start(Common#REGISTRY, IO_BLOCK, Common#RX_PIN, Common#TX_PIN, MODE, BAUD, TRUE)
#ifdef NO_HMI
#else    
  ' set up the proxy parameters
  HMI.ProxySetup(IO_BLOCK, Proxy_Lock, Server_CPU)
#endif  
#endif

#ifdef START_GRAPHICS
  CGI_DATA := ALLOC - CGI#BYTE_SIZE
  CGI_DATA := 128 * (CGI_DATA / 128) ' must align to a 128 byte boundary
  ALLOC    := CGI_DATA
  ' start the selected CGI Plugin (and graphics drivers)
  if ok
    ok := CGI.Start(Common#REGISTRY, CGI_DATA)
#endif

#ifdef NO_HMI
#else
  ' allocate data block for HMI
  HMI_DATA := ALLOC - 4 * HMI#DATASIZE
  ALLOC    := HMI_DATA
  ' start the selected HMI Plugin (and keyboard, mouse & screen drivers)
  if ok
    ok := HMI.Start(Common#REGISTRY, HMI_DATA, START_KBD, START_MOUSE, START_SCREEN)
#endif

  ' start additional plugins here
#ifdef NO_FP
#elseifdef NO_FLOAT
#else
  if ok
    ok := Float_A.Start(Common#REGISTRY)
#ifdef libmb
  if ok
    ok := Float_B.Start(Common#REGISTRY)
#endif
#endif

#ifdef CLOCK
  if ok
    ' start the real-time clock - the second parameter controls the debug
    ' LED (on the Hydra) : 0 => LED off, 1 => LED toggles once per second
    ok := RTC.Start(Common#REGISTRY, 0)
#endif

#ifdef NEED_SD
  if ok
    ' start the SD Card Support plugin
#ifdef PROXY_SD
    ok := SD.Start(Common#REGISTRY, Common#PROXY_XFER, IO_BLOCK, Proxy_Lock, Server_CPU)
#else  
    ok := SD.Start(Common#REGISTRY)
#endif    
#endif

  ' all plugins now loaded - start the LMM by giving it the address
  ' of the top of RAM (to use as the initial SP) and also the base 
  ' address of the Catalina program to execute

  if ok
    Common.SendInitializationData(cog, base, ALLOC)

  
PUB startDebug(base)

  stopDebug
  PreStart
  DebugCog := StartCog + 1
  PostStart(DebugCog - 1, base)

PUB initDebug(dbgParamPtr)

  LONG[dbgParamPtr][-2] := _DbgParam
  LONG[dbgParamPtr][-1] := _DbgParamPtr
  LONG[dbgParamPtr][1]  := @debugEntry
  LONG[dbgParamPtr][2]  := _DbgTemp
  _DbgParam := dbgParamPtr


PUB stopDebug

  if DebugCog
    cogstop(DebugCog~ - 1)

'PUB LMM_Base : addr
'  addr := Catalina.Base

DAT
''
''The debug kernel requires 36 longs of Cog memory including the jmp statement to give control to
''the debugger.
''
                        org     0
debugEntry              jmp     #dbg_init               'initialize kernel, then start debugger
'
'PLACE YOUR CODE TO DEBUG HERE EXCEPT THE UNINITIALIZED VARIABLES!
'PLACE YOUR UNINITIALIZED VARIABLES AT THE END OF THIS FILE!
'I RECOMMEND TO PLACE ALL YOUR VARIABLES AT THE END OF THIS FILE EXCEPT THOSE ARE EVENTUALLY
'LOCATED BETWEEN ASSEMBLY INSTRUCTIONS!
'
'-----------------------------------------------------------------------------------------
'
entry
'
'--------------------- Start of LMM Kernel ------------------------------------
'
   jmp #INIT            '$00
'
'------------------------------------------------------------------------------
'  jmp #INIT            '$01 remove this line when using POD, since it
                        '    adds a jump instruction before the program
'------------------------------------------------------------------------------
'
INIT    jmp #lmm_init   '$02
LODL    jmp #load_l     '$03
LODA    jmp #load_a     '$04
LODF    jmp #fp_ind     '$05
PSHL    jmp #push_l     '$06
PSHB    jmp #push_b     '$07
CPYB    jmp #copy_b     '$08
NEWF    jmp #new_fp     '$09
RETF    jmp #fp_ret     '$0a
CALA    jmp #f_call     '$0b
RETN    jmp #f_ret      '$0c
CALI    jmp #f_cali     '$0d
JMPA    jmp #f_jump     '$0e
JMPI    jmp #f_jmpi     '$0f
DIVS    jmp #f_d32s     '$10
DIVU    jmp #f_d32u     '$11
MULT    jmp #f_m32      '$12
BR_Z    jmp #bra_z      '$13
BRNZ    jmp #bra_nz     '$14
BRAE    jmp #bra_ae     '$15
BR_A    jmp #bra_a      '$16
BRBE    jmp #bra_be     '$17
BR_B    jmp #bra_b      '$18
SYSP    jmp #plugin     '$19
PSHA    jmp #push_a     '$1a
FADD    jmp #flt_add    '$1b
FSUB    jmp #flt_sub    '$1c
FMUL    jmp #flt_mul    '$1d
FDIV    jmp #flt_div    '$1e
FCMP    jmp #flt_cmp    '$1f
FLIN    jmp #flt_int    '$20
INFL    jmp #int_flt    '$21
PSHM    jmp #push_m     '$22
POPM    jmp #pop_m      '$23
PSHF    jmp #push_i     '$24
RLNG    jmp #rd_long    '$25
RWRD    jmp #rd_word    '$26
RBYT    jmp #rd_byte    '$27
WLNG    jmp #wr_long    '$28
WWRD    jmp #wr_word    '$29
WBYT    jmp #wr_byte    '$2a
'                           
PC      long 0          '$2b
SP      long 0          '$2c
FP      long 0          '$2d
RI      long 0          '$2e
BC      long 0          '$2f
BA      long 0          '$30
BZ      long 0          '$31
CS      long 0          '$32
'
r0      long 0          '$33
r1      long 0          '$34
r2      long 0          '$35
r3      long 0          '$36
r4      long 0          '$37
r5      long 0          '$38
r6      long 0          '$39
r7      long 0          '$3a
r8      long 0          '$3b
r9      long 0          '$3c
r10     long 0          '$3d
r11     long 0          '$3e
r12     long 0          '$3f
r13     long 0          '$40
r14     long 0          '$41
r15     long 0          '$42
r16     long 0          '$43
r17     long 0          '$44
r18     long 0          '$45
r19     long 0          '$46
r20     long 0          '$47
r21     long 0          '$48
r22     long 0          '$49
r23     long 0          '$4a
'
Bit31   long  $80000000 '$4b
all_1s  long  $ffffffff '$4c
cviu_m1 long  $000000ff '$4d
cviu_m2 long  $0000ffff '$4e
top8    long  $ff000000 '$4f   ' top 8 bits bitmask
low24   long  $00ffffff '$50   ' low 24 bits bitmask
'
init_BZ long  $00000000 '$51   ' end of code / start of heap
init_PC long  $00000000 '$52   ' the initial PC
'
' dbg_init - initialize VM, then start debugger
' (replaces lmm_init, which  now just starts executing)
'
dbg_init
        cogid   r0              ' convert ...
        shl     r0,#2           ' ... my cog id ...
        add     r0,par          ' ... to my registration addr
        rdlong  r0,r0           ' get my request block addr
wait
        rdlong  BA,r0 wz        ' Wait till we are given the base address ...
  if_z  jmp     #wait           ' .. (required to relocate segments)
        add     r0,#4           ' Load our initial SP ...
        rdlong  SP,r0           ' ... from the request block
        mov     BZ,#init_BZ<<2+8 ' calculate ...
        add     BZ,BA           ' ... pointer to initial BZ
        mov     PC,BZ           ' load ...
        add     PC,#4           ' ... initial  ...
        rdlong  PC,PC           ' ... PC and ...
        rdlong  BZ,BZ           ' ... BZ and ...
        call    #Reloc          ' ... relocate segments 
        jmp     #debugger       'start debugger
        nop
'
' LMM_next - increment the PC then execute the instruction pointed to by the new PC
'        
LMM_next
        add    PC,#4
        
lmm_init
'
' LMM_loop - Fetch an instruction from HUB RAM to COG RAM and execute it.
' The basic LMM loop is unrolled 4 times to increase speed.
'
LMM_loop
        rdlong LMM_1,PC
        add    PC,#4
LMM_1
        nop
        rdlong LMM_2,PC
        add    PC,#4
LMM_2
        nop
        rdlong LMM_3,PC
        add    PC,#4
LMM_3
        nop
        rdlong LMM_4,PC
        add    PC,#4
LMM_4
        nop
        jmp    #LMM_loop

'
' Reloc - move segments as indicated by the BA. We move all Catalina
'         data (i.e. from BA to BZ) to start at location zero, but do 
'         not overwrite the first 16 bytes of RAM, since they contain
'         special values (clock freq etc).
'
Reloc
        mov   BC,BZ         ' length of data to be relocated ...
        add   BC,#3         ' round up ...
        andn  BC,#3         ' ... to a multiple of 4 bytes
        sub   BC,#$10       ' omit first 16 bytes
        mov   t2,#$10       ' destination is byte 16
        mov   t1,BA         ' source is BA ...
        add   t1,#$10       ' ... plus 16 bytes 
reloc_loop
       tjz    BC,#Reloc_ret ' no more to copy
       rdlong t3,t1         ' read from src to t3
       wrlong t3,t2         ' write t3 to dst
       add    t1,#4         ' increment source
       add    t2,#4         ' increment destination
       sub    BC,#4         ' decrement count ...
       jmp    #reloc_loop   ' ... and keep copying
Reloc_ret
       ret        
'
'load_l - load the long stored at the PC into RI, incrementing the PC
'         NOTE: load_l is for long constants - use 'load_a' for addresses
'
load_l
       rdlong RI,PC         ' load the long value
       jmp    #LMM_next     ' increment PC, execute next instruction
'
'load_a - load the long address at the PC into RI, incrementing the PC
'         NOTE: load_a is for addresses - use 'load_l' for constants
'
load_a
       rdlong RI,PC         ' load the address
       jmp    #LMM_next     ' increment PC, execute next instruction
'
' fp_ind - load the FP into RI, then add the long at the PC to result, incrementing the PC
'
fp_ind
       rdlong RI,PC         ' load the long value
       adds   RI,FP         ' add the frame pointer
       jmp    #LMM_next     ' increment PC, execute next instruction
'
' push_l - push a long value in RI onto the stack, decrementing the SP
'
push_l
       sub    SP,#4         ' decrement SP
       wrlong RI,SP         ' save value on stack
       jmp    #LMM_loop     ' increment PC, execute next instruction
'
' push_a - push a value whose address is specified indirectly in a global
'          variable onto the stack
'
push_a
       rdlong RI,PC         ' load the address
       jmp    #push_x       ' use the result as an address of the value to push
'
' push_i - push a value whose address is specified indirectly in a frame variable
'          onto the stack
'
push_i
       rdlong RI,PC         ' load the long value
       adds   RI,FP         ' add the frame pointer
push_x
       rdlong RI,RI         ' read the value at that address
       sub    SP,#4         ' decrement SP
       wrlong RI,SP         ' save value on stack
       jmp    #LMM_next     ' increment PC, execute next instruction
'
' push_m - push many registers (specified in RI) onto the stack,
' decrementing the SP before each one
'
push_m
       rdlong RI,PC         ' load the long value specifying the registers to push
       movd   :push_x,#r0   ' start with r0
       mov    t1,#24        ' we will push up to 24 registers
:push_nxt
       shr    RI,#1 wc,wz   ' save rx?
 if_c  sub    SP,#4         ' if so ...
:push_x                     ' ... save ...
 if_c  wrlong 0,SP          ' ... the register
 if_z  jmp    #LMM_next     ' no more to save
       add    :push_x,dlsb  ' point to next register
       djnz   t1,#:push_nxt ' continue till all 24 registers checked
       jmp    #LMM_next     ' increment PC, execute next instruction
        
'
' pop_m - pop many registers (specified in RI) from the stack,
' incremening the SP after each one
'
'
pop_m
       rdlong RI,PC         ' load the long value specifying the registers to push
       movd   :pop_x,#r23   ' start with ...
       shl    RI,#(32-24)   ' ... r23         
       mov    t1,#24        ' we will pop up to 24 registers
:pop_nxt
       shl    RI,#1 wc      ' load rx?
:pop_x                      ' if so ...
 if_c  rdlong 0,SP          ' ... load ...
 if_c  add    SP,#4         ' ... the register
       sub    :pop_x,dlsb   ' point to previous register
       djnz   t1,#:pop_nxt  ' continue till all 24 registers checked
       jmp    #LMM_next     ' increment PC, execute next instruction

'
' push_b - push a structure (size in bytes at the PC) pointed to by R0
'          onto the stack, decrementing the SP. Note that the source
'          and destination data MUST be long aligned - but this is true
'          for all normal structures.
'
push_b
       rdlong BC,PC         ' load the byte count
       add    PC,#4         ' increment the PC
       add    BC,#3         ' round up BC ...
       andn   BC,#3         ' ... to be a multiple of 4
       sub    SP,BC         ' decrement SP by rounded up size
       mov    t1,R0         ' source is in R0
       mov    t2,SP         ' destination is SP
       jmp    #copy_longs   ' do the copy
'
' copy_b - copy a structure (size in bytes at the PC)
'          from the address in R1 to the address in R0
'
copy_b
       rdlong BC,PC         ' load the byte count
       add    PC,#4         ' increment the PC
       add    BC,#3         ' round up BC ...
       andn   BC,#3         ' ... to be a multiple of 4
       mov    t1,R1         ' source is in R1
       mov    t2,R0         ' destination is in R0
'
copy_longs
       tjz    BC,#LMM_Loop  ' no more to copy
       rdlong t3,t1         ' read from src to t3
       wrlong t3,t2         ' write t3 to dst
       add    t1,#4         ' increment source
       add    t2,#4         ' increment destination
       sub    BC,#4         ' decrement count ...
       jmp    #copy_longs   ' ... and keep copying
'
' new_fp - save current frame pointer and set up a new frame pointer
' also calculate what SP was before data was pushed and save it in BC
'
new_fp
       sub    SP,#4         ' decrement the stack pointer
       wrlong FP,SP         ' save FP to stack
       mov    FP,SP         ' set up new FP
       add    BC,#8         ' calculate what SP was ...
       add    BC,FP         ' ... before arguments were pushed
       jmp    #LMM_loop     ' execute the next instruction
'
' fp_ret - pop current frame, restore previous frame and stack pointers,
'          then restore the previous program counter
'
' f_ret -  just restore the previous program counter
'
fp_ret
       rdlong FP,SP         ' restore previous FP
       add    SP,#4         ' increment the SP
f_ret
       rdlong PC,SP         ' read the PC
       add    SP,#4         ' increment the SP
       jmp    #LMM_loop     ' execute the next instruction
'
' f_call - call the routine at the address pointed to by the PC (increment the PC)
' f_cali - call the routine whose address is in RI
'
f_call
       rdlong RI,PC         ' get the address to call
       add    PC,#4         ' increment the PC (this is the return address)
f_cali
       sub    SP,#4         ' decrement the SP
       wrlong PC,SP         ' save current PC to stack
       jmp    #f_jmpi       ' jump to location in RI
'
' f_jump - jump to the location at the address pointed to by PC (increment the PC)
'
f_jump
       rdlong PC,PC         ' get the address to jump to
       jmp    #LMM_loop     ' execute next instruction
'
' f_jmpi - jump to the location whose address in RI
'
f_jmpi
       mov    PC,RI         ' get the address to jump to
       jmp    #LMM_loop     ' execute next instruction
'
' rd_long/rd_word/rd_byte : Read long/word/byte from HUB memory at address in RI into BC.
'
' On entry:
'    RI : address to read
' On exit:
'    BC : long/word/byte read
'
rd_long
        rdlong  BC,RI
        jmp     #LMM_Next
rd_word
        rdword  BC,RI
        jmp     #LMM_Next
rd_byte
        rdbyte  BC,RI
        jmp     #LMM_Next
'
' wr_long/wr_word/wr_byte : Write long/word/byte to HUB memory at address in RI into BC.
'
' On entry:
'    RI : address to write
'    BC : long/word/byte to write
' On exit:
'    (none)
'
wr_long
        wrlong  BC,RI
        jmp     #LMM_Next
wr_word
        wrword  BC,RI
        jmp     #LMM_Next
wr_byte
        wrbyte  BC,RI
        jmp     #LMM_Next
'
'fd_32s - Signed 32 bit division
'         Divisor --> r1
'         Dividend--> r0
'         Result --> Quotient in r0
'                    Remainder in r1
'
#ifdef OLD_DIVISION

' use original division algorithm

f_d32s
        mov t1,#32
        mov t2,#0           ' tmp2 is temp register to hold the remainder
        mov ftemp,r1
        shr ftemp,#31 WZ    ' get the divisor sign (ftemp)
 if_nz  sub t2,r1           ' divisor = -divisor if < 0
 if_nz  mov r1,t2
        mov ftmp2,r0
        shr ftmp2,#31 WZ    ' get the dividend sign (ftmp2)
        mov t2,#0
 if_nz  sub t2,r0           ' dividend = -dividend if < 0
 if_nz  mov r0,t2
        mov t2,#0
        xor ftemp,ftmp2     ' sign of result (ftemp)
:up
        shl r0,#1 WC
        rcl t2,#1 WC
        cmp r1,t2 WC,WZ
 if_a   jmp #:down1
        sub t2,r1
        add r0,#1
:down1
        sub t1,#1 WZ
 if_ne  jmp #:up
        mov r1,t2
        mov t2,#0
        cmp ftmp2,#0 WZ
 if_nz  sub t2,r1           ' adjust the sign of remainder
 if_nz  mov r1,t2
        mov t2,#0
        cmp ftemp,#0   WZ
 if_nz  sub t2,r0
 if_nz  mov r0,t2           ' adjust the sign of quotient
        jmp #LMM_loop

#else

' use Pullmoll's new division algorithm

f_d32s
        mov  t1,#32
        mov  t2,#0           ' tmp2 is temp register to hold the remainder
        mov  ftemp,r1
        mov  ftmp2,r0
        abs  r1, r1
        abs  r0, r0
        xor  ftemp, ftmp2
:loop
        shl  r0,#1 WC
        rcl  t2,#1 WC
        cmpsub t2,r1 WC,WZ
 if_be add  r0,#1
        djnz t1,#:loop
        mov  r1,t2
        mov  ftmp2,ftmp2 WC  ' get bit #31 into carry
        negc r1,r1                   ' adjust the sign of the remainder
        mov  ftemp,ftemp WC  ' get bit #31 into carry
        negc r0,r0                   ' adjust the sign of the result
        jmp  #LMM_loop
        
#endif
        
'
'f_d32u - Unsigned 32 bit division
'         Divisor : r1
'         Dividend : r0
'         Result:
'             Quotient in r0
'             Remainder in r1

f_d32u
        mov ftemp,#32
        mov ftmp2, #0
:up2
        shl r0,#1       WC
        rcl ftmp2,#1    WC
        cmp r1,ftmp2    WC,WZ
 if_a   jmp #:down
        sub ftmp2,r1
        add r0,#1
:down
        sub ftemp, #1   WZ
 if_ne  jmp #:up2
        mov r1,ftmp2
        jmp #LMM_loop
'
'f_m32 - multiplication
'        r0 : 1st operand (32 bit)
'        r1 : 2nd operand (32 bit)
'        Result:
'           Product in r0 (<= 32 bit)
'
f_m32
        mov ftemp,#0
:start
        cmp r0,#0       WZ
 if_e   jmp #:down3
        shr r0,#1       WC
 if_ae  jmp #:down2
        add ftemp,r1    WC
:down2
        shl r1,#1       WC
        jmp #:start
:down3
        mov r0,ftemp
        jmp #LMM_loop
'
' bra_xx - branch if condition is true to the address at the PC,
'          otherwise, just increment the PC by 4
'
'
bra_z
  if_z  jmp    #f_jump       ' if condition true, branch is equiv to jump
        jmp    #LMM_next     ' increment PC, execute next instruction
bra_nz
 if_nz  jmp    #f_jump       ' if condition true, branch is equiv to jump
        jmp    #LMM_next     ' increment PC, execute next instruction
bra_ae
 if_ae  jmp    #f_jump       ' if condition true, branch is equiv to jump
        jmp    #LMM_next     ' increment PC, execute next instruction
bra_a
 if_a   jmp    #f_jump       ' if condition true, branch is equiv to jump
        jmp    #LMM_next     ' increment PC, execute next instruction
bra_be
 if_be  jmp    #f_jump       ' if condition true, branch is equiv to jump
        jmp    #LMM_next     ' increment PC, execute next instruction
bra_b
 if_b   jmp    #f_jump       ' if condition true, branch is equiv to jump
        jmp    #LMM_next     ' increment PC, execute next instruction

flt_add
        call   #_FAdd        ' r0 = r0 + r1
        jmp    #LMM_loop     ' execute next instruction
flt_sub
        call   #_FSub        ' r0 = r0 - r1
        jmp    #LMM_loop     ' execute next instruction
flt_mul
        call   #_FMul        ' r0 = r0 * r1
        jmp    #LMM_loop     ' execute next instruction
flt_div
        call   #_FDiv        ' r0 = r0 / r1
        jmp    #LMM_loop     ' execute next instruction
flt_cmp
        jmp    #_FCmp        ' set Z and C flags based on r0 = R1
flt_int
        call   #_FFloat      ' r0 = float(r0)
        jmp    #LMM_loop     ' execute next instruction
int_flt
        call   #_FTrunc      ' r0 = trunc(r0)
        jmp    #LMM_loop     ' execute next instruction

'
' plugin - call a plugin
' On entry:
'       R3 = code ($80 + cog id, or plugin type)
'       R2 = data (optional pointer)
'
' On exit:
'       R0 = result
'
'
plugin
        mov     r0,r3           ' get plugin type
        test    r0,#$80 wz      ' request by type?
 if_z   jmp     #ptype          ' yes - must find plugin of this type
        and     r0,#$7f         ' no - code is cog id
        jmp     #pid            ' invoke plugin by cog id
ptype

        mov     t1,par          ' no - point to registry
        mov     t2,#0           ' start at cog 0
plgt1
'
' TODO - should remember plugins once found by type!
'
        cmp     t2,#8 wc,wz     ' run out of plugins?
 if_ae  jmp     #plugerr        ' yes - no such plugin
        rdlong  t3,t1           ' no - check next plugin type
        shr     t3,#24          ' is it ...
        cmp     t3,r0 wz        ' ... the type what we wanted?
 if_z   jmp     #plgt2          ' yes - use this plugin
        add     t2,#1           ' no ...
        add     t1,#4           ' ... check ...
        jmp     #plgt1          ' ... next cog
plgt2
        mov     r0,t2           ' use the cog where we found the plugin
pid
        shl     r0,#2           ' multiply plugin (cog) id by 4 ...
        add     r0,par          ' add registry base to get registry entry
        rdlong  r0,r0           ' get request block from registry
        test    r0,top8 wz      ' plugin registered?
 if_z   jmp     #plugerr        ' no - error 255
        and     r0,low24        ' yes - write request ...
        wrlong  r2,r0           ' ... to request block
loop    rdlong  r3,r0   wz      ' wait till ...
 if_nz  jmp     #loop           ' ... request completed
        add     r0,#4           ' get results ...
        rdlong  r0,r0           ' ... from request block
        jmp     #LMM_loop       'done

plugerr neg     r0,#1           ' plugin not registered
        jmp     #LMM_loop

'------------------------------------------------------------------------------
' Float32 Assembly language routines
'------------------------------------------------------------------------------

_FAdd                   mov     t2, #FAddCmd>>24
                        jmp     #send
_FSub                   mov     t2, #FSubCmd>>24
                        jmp     #send
_FMul                   mov     t2, #FMulCmd>>24
                        jmp     #send
_FDiv                   mov     t2, #FDivCmd>>24
                        jmp     #send
_FFloat                 mov     t2, #FFLoatCmd>>24
                        jmp     #send
_FTrunc                 mov     t2, #FTruncCmd>>24

send                    call    #sendCmd
              
_FAdd_ret
_FSub_ret
_FMul_ret
_FDiv_ret
_FFloat_ret
_FTrunc_ret
                        ret

'------------------------------------------------------------------------------
' sendCmd - send a command to Float A for execution
' input:   t2           command
'          r0          32-bit floating point value
'          r1          32-bit floating point value 
' output:  r0          32-bit floating point result
'------------------------------------------------------------------------------

sendCmd
        shl     t2, #24         ' command to send to Float32A  
        mov     t1, xfer        ' get pointer to xfer block 
        or      t2, t1          ' construct request

        wrlong  r0, t1          ' write first argument to xfer block
        add     t1, #4
        wrlong  r1, t1          ' write second argument to xfer block

        mov     r0,#LMM_FLA     ' plugin type we want 

        mov     ftemp,par       ' point to registry
        mov     ftmp2,#0        ' start at cog 0
send1
        cmp     ftmp2,#8 wc,wz  ' run out of plugins?
 if_ae  jmp     #sendErr        ' yes - no such plugin
        rdlong  ftmp3,ftemp     ' no - check next plugin type
        shr     ftmp3,#24       ' is it ...
        cmp     ftmp3,r0 wz     ' ... the type what we wanted?
 if_z   jmp     #send2          ' yes - use this plugin
        add     ftmp2,#1        ' no ...
        add     ftemp,#4        ' ... check ...
        jmp     #send1          ' ... next cog
send2
        mov     r0,ftmp2        ' use the cog where we found the plugin
        shl     r0,#2           ' multiply plugin (cog) id by 4 ...
        add     r0,par          ' add registry base to get registry entry
        rdlong  r0,r0           ' get request block from registry
        test    r0,top8 wz      ' plugin registered?
 if_z   jmp     #sendErr        ' no - return NaN
        and     r0,low24        ' yes - write request ...
        wrlong  t2,r0           ' ... to request block
loop2   rdlong  t2,r0   wz      ' wait till ...
 if_nz  jmp     #loop2          ' ... request completed
        add     r0,#4           ' get results ...
        rdlong  r0,r0           ' ... from request block
        jmp     #sendCmd_ret
sendErr
        mov     r0, NaN         ' return NaN on any error
sendCmd_ret
        ret
        
'--------------------- End of LMM Kernel --------------------------------------

'--------------------- Start of Float32 Components ----------------------------

{{

+------------------------------------------------------------------------------------------------------------------------------+
?                TERMS OF USE: MIT License  (Float32 Components only - i.e. excludes LMM Kernel)                               ?                                                            
+------------------------------------------------------------------------------------------------------------------------------?
?Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    ? 
?files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    ?
?modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software?
?is furnished to do so, subject to the following conditions:                                                                   ?
?                                                                                                                              ?
?The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.?
?                                                                                                                              ?
?THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          ?
?WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         ?
?COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   ?
?ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         ?
+------------------------------------------------------------------------------------------------------------------------------+
}}

'------------------------------------------------------------------------------
' _FCmp    set Z and C flags for r0 - r1
' changes: t1
'------------------------------------------------------------------------------

_FCmp                   mov     t1, r0               ' compare signs
                        xor     t1, r1
                        and     t1, Bit31 wz
          if_z          jmp     #:cmp1                  ' same, then compare magnitude

                        mov     t1, r0               ' check for +0 or -0
                        or      t1, r1
                        andn    t1, Bit31 wz,wc
          if_z          jmp     #LMM_Loop

                        test    r0, Bit31 wc         ' compare signs
                        jmp     #LMM_Loop

:cmp1                   test    r0, Bit31 wz         ' check signs
          if_nz         jmp     #:cmp2
                        cmp     r0, r1 wz,wc
                        jmp     #LMM_Loop
:cmp2                   cmp     r1, r0 wz,wc      ' reverse test if negative
                        jmp     #LMM_Loop



'-------------------- Float32 constant values ----------------------------------

NaN                     long    $7FFF_FFFF
Minus23                 long    -23
Mask23                  long    $007F_FFFF
Bit29                   long    $2000_0000
dlsb                    long    1 << 9
xfer                    long    FLOAT_XFER              ' transfer to Float32A



'
'==================================================================================================
{
Code of kernel to debug your assembly program. The debug kernel requires 36 longs of Cog memory
including a jmp statement to give control to the debugger.
}
getData                 movs    :readData,_DbgAddr
                        nop
:readData               mov     _DbgTemp,0-0
getData2                wrlong  _DbgTemp,_DbgParamPtr

'-----------------------
debugger                mov     _DbgTemp,#0
                        wrlong  _DbgTemp,_DbgParam      'send ready
                        mov     _DbgParamPtr,_DbgParam
                        add     _DbgParamPtr,#4         'point to DebugAddr
:dbgWaitCmd             rdlong  _DbgTemp,_DbgParam
                        tjz     _DbgTemp,#:dbgWaitCmd
                        rdlong  _DbgAddr,_DbgParamPtr
                        add     _DbgTemp,#dbgCmdTable - 1
                        movs    :dbgJmpCmd,_DbgTemp
                        add     _DbgParamPtr,#4         'point to DebugData
:dbgJmpCmd              jmp     #0-0                    'this is also used to run the debugged code
dbgCmdTable             jmp     #getMainData
                        jmp     #setMainData
                        jmp     #getData
                        jmp     #setData

'-----------------------
getFlags                muxnz   _DbgTemp,#FLAG_NZ
                        muxc    _DbgTemp,#FLAG_C
                        jmp     #getData2

'-----------------------
setData                 movd    :writeData,_DbgAddr
                        rdlong  _DbgTemp,_DbgParamPtr
:writeData              mov     0-0,_DbgTemp
                        jmp     #debugger

'-----------------------
getMainData             rdlong  _DbgTemp,_DbgAddr
                        jmp     #getData2

'-----------------------
setMainData             rdlong  _DbgTemp,_DbgParamPtr
                        wrlong  _DbgTemp,_DbgAddr
                        jmp     #debugger

'#######################################################
'variables used by debugger
_DbgParam               jmpret  _DbgTemp,#getData2      'this points always to the first parameter like register par
_DbgParamPtr            long    dbgCmdTable - 1         'pointer to access parameters
_DbgAddr                long    0                       'debugged address
_DbgTemp                long    _DbgVariables
_DbgVariables
'==================================================================================================
'
'PLACE YOUR UNINITIALIZED VARIABLES HERE!
'
'-------------------- Float32 local variables ---------------------------------

t1                      res     1                       ' temporary values
t2                      res     1
t3                      res     1

status                  res     1                       ' last compare status

flagA                   res     1
expA                    res     1
manA                    res     1

flagB                   res     1
expB                    res     1
manB                    res     1

'-------------------- End of Float32 ------------------------------------------
'
' temporary storage used in mul & div (and sendCmd in debug target) calculations
'
ftemp   res 1
ftmp2   res 1
ftmp3   res 1

'--------------------- End of Float32 Components ------------------------------
'
              fit       $1eb                    ' last 5 longs reserved for debug overlay vectors
'              
              org       $1eb
DEBUG_VECTORS long      0,0,0,0,0
              fit       $1f0                    ' max size
